--import "cmds"
if not import then
	dofile "import.lua"
end
import "cmds"
import "string"
import "table"
import "pickle"

--Schema
--[[Class]]User = {
	id = 1,
	schema = {
		[0] = "id",
		"username",
		"password",
		"usgn",
		"rank",
	},
	users = {},
	cache = {},
	usecache = true,
	players = {},
	path = "user.db"
}
setmetatable(User, {__call=function(class, ...)
	local args = {...}
	if type(args[1]) == "table" then
		args = args[1]
	end
	local object = {}

	--ID
	object.id = class.id
	class.id = class.id + 1

	for offset, key in ipairs(class.schema) do
		object[key] = args[offset] or args[key] or false -- Delete last switch to disable
	end

	class.users[object.id] = object

	return object
end})

User.users = pickle.load(User.path) or {}
User.id = #User.users + 1
User.cache = pickle.load("sys/lua/userscache.db") or {}


function User.resolve(key)
--~ 	if User.cache[key] then
--~ 		return User.cache[key]
--~ 	end
	for id, user in ipairs(User.users) do
		if user.username == key or user.usgn == tonumber(key) then
			setmetatable(user, {__newindex = function(self, k, v)
									rawset(self, k, v or false);
									pickle.dump(User.users, User.path)
								end,
								__call = function()
									pickle.dump(User.users, User.path)
								end})
			return user
		end
	end
end

function User.register(username, password, usgn)
	if tonumber(username) then return end
	if not usgn or usgn == 0 then usgn = nil end
	if User.resolve(username) then return end
	local user = User{username = username, password = password, usgn = usgn, rank = {user = true}}
	pickle.dump(User.users, User.path)
	setmetatable(user, {__newindex = function(self, k, v) rawset(self, k, v or false); pickle.dump(User.users, User.path) end,
								__call = function()
									pickle.dump(User.users, User.path)
								end})
	return user
end

function User.login(id, username, password)
	local user = User.resolve(username)
	if not username then
		username = player(id, "usgn")
		if user then
			password = user.password
		end
	end
	if not user then
		return msg2(id, "Failed to login: User '%s' does not exist"%username)
	end
	if password == user.password then
		User.players[id] = user
		return user
	else
		return msg2(id, "Failed to login: incorrect password for user '%s'"%username)
	end
end

function User.logout(id)
	User.players[id] = nil
end

function User.find_as_player(username)
	for id, user in pairs(User.players) do
		if user.username == username then return id end
	end
end

function act.login(id, username, password)
	__help__ =
	[[
		-Description:
		... Login to the user system
		-Usage:
		... !login <username> <password>
		-Example:
		... !login dummy 123456
	]]
	local user = User.login(id, username, password)
	if user then msg2(id, "Successfully logged in as '%s'"%username) end
end

function act.register(id, username, password)
	__help__ =
	[[
		-Description:
		... Register your information into the user system
		-Usage:
		... !register <username> <password>
		-Example:
		... !register dummy 123456
	]]
	local user = User.register(username, password, player(id, "usgn"))
	if user then msg2(id, "Successfully registered as '%s'"%username) end
	User.login(id, username, password)
end

function act.logout(id)
	__help__ =
	[[
		-Description:
		... Logs you out of the user system
		-Usage:
		... !logout
	]]
	if User.players[id] then
		msg2(id, "Successfully logged out from the account '%s'."%User.players[id].username)
		User.logout(id)
	else
		msg2(id, "You are not currently logged in.")
	end
end


adm = cmds.category("admin","@")
adm.rank = "user"
function adm:preprocessing(id, cmd, text, this)
	local user = User.players[id]
	if user then
		this.user = user
		if this.user.rank[this.rank] then
			return this
		else
			local sep = ", "
			local stack = {}
			for _,v in pairs(this.user.rank) do table.insert(stack,"'%s'"%_) end
			table.sort(stack)
			msg2(id, "You do not have sufficient privilege to use the command '%s'"%cmd)
			msg2(id, "You need a rank of '%s', you are currently %s"%{this.rank, sep:join(stack)})
			return false
		end
	else
		msg2(id, "You must be logged in to use the command '%s'"%cmd)
		return false
	end
end

function adm.usgn(this, id)
	__help__ =
	[[
		-Description:
		... Change the USGN of the current user
		-Usage:
		... @usgn
	]]
	this.user.usgn = tonumber(player(id, "usgn"))
	this.user() -- Save
	msg2(id, "USGN successfully changed to '%s'"%(tonumber(usgn) or "0"))
end

function adm.username(this, id, username)
	__help__ =
	[[
		-Description:
		... Change the username of the current user
		-Usage:
		... @username <username>
	]]
	if User.resolve(username) then
		return msg2(id, "Cannot change username to '%s'"%username)
	end
	this.user.username = username
	this.user()
	msg2(id, "Username successfully changed to '%s'"%username)
end

function adm.password(this, id, password)
	__help__ =
	[[
		-Description:
		... Change the password of the current user
		-Usage:
		... @password <password>
	]]
	this.user.password = password
	this.user()
	msg2(id, "Password successfully changed.")
end

function adm.who(id, ...)
	__help__ =
	[[
		-Description:
		... Displays your or other's account information
		-Usage:
		... @who -> Displays your own user information
		... @who <id> -> Displays Player ID's user information
	]]
	local args = {...}
	local target = tonumber(args[1]) or id
	local user = User.players[target]
	if user then
		msg2(id, "%s's username is '%s'"%{player(target, "name"), user.username})
	else
		msg2(id, "%s is not currently logged in"%(player(target,"name") or "NaN"))
	end
end

function adm.auth.superadmin(id, this, username, rank1, ...)
	__help__ =
	[[
		-Description:
		... Alters a player's privileges, the added ranks are appended to the
		... user's previous ranks.
		-Usage:
		... @auth <username> <rank> <[rank2 [rank3 [...]\]\]>
		-Example:
		... @auth dummy admin
		... @auth dummy user admin
	]]
	local ranks = {rank1, ...}
	local user = User.resolve(username)
	for _,rank in ipairs(ranks) do user.rank[rank] = true end
	user()
	local target = User.find_as_player(username)
	msg2(id, "Succesfully altered %s's privileges."%username)
	if target then
		local sep = ", "
		msg2(target, "Your ranks have been elevated to '%s'"%sep:join(ranks))
		User.players[target] = user
	end
end

function adm.authreset.superadmin(id, this, username, rank1, ...)
	__help__ =
	[[
		-Description:
		... Sets a user's ranks to only those defined by the arguments.
		-Usage:
		... @authreset <username> <rank> <[rank2 [rank3 [...]\]\]>
		-Example:
		... @authreset dummy admin
		... @authreset dummy user admin
	]]
	local ranks = {rank1, ...}
	local user = User.resolve(username)
	user.rank = {}
	for _,rank in ipairs(ranks) do user.rank[rank] = true end
	user()
	local target = User.find_as_player(username)
	msg2(id, "Succesfully altered %s's privileges."%username)
	if target then
		local sep = ", "
		msg2(target, "Your ranks have been set to '%s'"%sep:join(ranks))
		User.players[target] = user
	end
end

if addhook then
	addhook("leave","hook_leave_auth")
end

function hook_leave_auth(id)
	User.logout(id)
end
