import "help" -- Try
import "string"
import "table"

local function copy(t)
	local tab = {}
	for _,v in ipairs(t) do
		table.insert(tab, v)
	end
	return tab
end

local function ReadInt(str)
	local t = {}
	for w in str:gmatch(".") do table.insert(t, w:byte()) end
	local b1 = t[1]
	local b2 = t[2]
	local b3 = t[3]
	local b4 = t[4]
	return b1+b2*256+b3*256*256+b4*256*256*256
end

local function num_args(func)
	local ok = pcall(function() string.dump(func) end)
	if not ok then return "?" end
	local dump = string.dump(func)
	--Set up the cursor
	local cursor = 1
	--Skip Header Stuff
	cursor = cursor + 12
	--Skip Source Name
	local offset = ReadInt(dump:sub(cursor, cursor + 4))
	cursor = cursor + 4
	local sourceName = dump:sub(cursor, cursor+offset)
	cursor = cursor + offset
	--Skip Line Defined
	cursor = cursor + 4
	--Skip Last Line Defined
	cursor = cursor + 4
	--Get the Upvalues
	local numUpvalues = dump:sub(cursor, cursor + 1):byte()
	cursor = cursor + 1
	--Get the Parameters
	local numParams = dump:sub(cursor, cursor + 1):byte()
	cursor = cursor + 1
	--Get the Variable-Args flag (whether there's a ... in the param)
	local varFlag = dump:sub(cursor, cursor + 1):byte()
	cursor = cursor + 1

	local ret = tostring(numParams)
	if varFlag > 1 then ret = ret .. "s" end

	return ret
end

local function get_args(func)
	local function _get_args(dump, cursor)
		--Set up the cursor
		if not cursor then
			cursor = 1
			--Skip Header Stuff
			cursor = cursor + 12
		end
		--Skip Source Name
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4
		local sourceName = dump:sub(cursor, cursor+offset)
		cursor = cursor + offset
		--Skip Line Defined
		cursor = cursor + 4
		--Skip Last Line Defined
		cursor = cursor + 4
		--Get the Upvalues
		local numUpvalues = dump:sub(cursor, cursor + 1):byte()
		cursor = cursor + 1
		--Get the Parameters
		local numParams = dump:sub(cursor, cursor + 1):byte()
		cursor = cursor + 1
		--Get the Variable-Args flag (whether there's a ... in the param)
		local varFlag = dump:sub(cursor, cursor + 1):byte()
		cursor = cursor + 1
		--Skip the Stacksizes
		cursor = cursor + 1
		--CODE:
		--~Skip the entire code...
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4
		cursor = cursor + offset*4
		--CONSTANTS:
		--~Skip the entire constant section...
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		--print(offset)
		cursor = cursor + 4
		for i=1,offset do
			local tp = dump:sub(cursor, cursor):byte()
			--print(tp)
			cursor = cursor + 1
			if tp == 1 then
				--Boolean
				--local off = ReadInt(dump:sub(cursor, cursor + 1))
				--print(dump:sub(cursor, cursor):byte())
				cursor = cursor + 1
			elseif tp == 3 then
				--Number
				local off = ReadInt(dump:sub(cursor, cursor + 8))
				--print(off)
				cursor = cursor + 8
			elseif tp == 4 then
				--String
				local off = ReadInt(dump:sub(cursor, cursor + 4))
				cursor = cursor + 4
				--print(dump:sub(cursor, cursor+off))
				cursor = cursor + off
			else
				-- Pass on nil
			end
		end
		--FUNCTIONS:
		--Reget...
		local offset = ReadInt(dump:sub(cursor, cursor + 4))

		cursor = cursor + 4
		for i=1, offset do
			local _
			_, cursor = _get_args(dump, cursor)
		end
		--LINES:
		--~Skip the entire Line section...
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4
		cursor = cursor + offset*4
		--LOCALS:
		--~Get Locals
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4
		local locals = {}
		local params = {}
		for i=1,offset do
			local off = ReadInt(dump:sub(cursor, cursor + 4))
			cursor = cursor + 4
			local varname = dump:sub(cursor, cursor+off)
			table.insert(locals,varname)
			if i <= numParams then table.insert(params,varname) end
			cursor = cursor + off
			cursor = cursor + 4
			cursor = cursor + 4
		end
		if varFlag > 1 then
			--table.insert(params, "...")
			params['...'] = true
		end
		--UPVALUES:
		--~Get Upvalues
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4
		for i=1,offset do
			local off = ReadInt(dump:sub(cursor, cursor + 4))
			cursor = cursor + 4
			cursor = cursor + off
		end
		return params, cursor
	end
	local ok = pcall(function() string.dump(func) end)
	if not ok then return {"..."} end
	local dump = string.dump(func)
	local ret = {_get_args(dump)}
	return ret[1]
end

local function get_help(func)
	local function _get_constants(dump, cursor)
		--Set up the cursor
		if not cursor then
			cursor = 1
			--Skip Header Stuff
			cursor = cursor + 12
		end
		--Skip Source Name
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4
		local sourceName = dump:sub(cursor, cursor+offset)
		cursor = cursor + offset
		--Skip Line Defined
		cursor = cursor + 4
		--Skip Last Line Defined
		cursor = cursor + 4
		--Get the Upvalues
		local numUpvalues = dump:sub(cursor, cursor + 1):byte()
		cursor = cursor + 1
		--Get the Parameters
		local numParams = dump:sub(cursor, cursor + 1):byte()
		cursor = cursor + 1
		--Get the Variable-Args flag (whether there's a ... in the param)
		local varFlag = dump:sub(cursor, cursor + 1):byte()
		cursor = cursor + 1
		--Skip the Stacksizes
		cursor = cursor + 1
		--CODE:
		--~Skip the entire code...
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4

		cursor = cursor + offset*4
		--CONSTANTS:
		--~Skip the entire constant section...
		local consts = {}
		local offset = ReadInt(dump:sub(cursor, cursor + 4))
		cursor = cursor + 4
		for i=1,offset do
			local tp = dump:sub(cursor, cursor):byte()
			cursor = cursor + 1
			if tp == 1 then
				--Boolean
				--local off = ReadInt(dump:sub(cursor, cursor + 4))
				cursor = cursor + 4
			elseif tp == 3 then
				--Number
				--local off = ReadInt(dump:sub(cursor, cursor + 8))
				cursor = cursor + 8
			elseif tp == 4 then
				--String
				local off = ReadInt(dump:sub(cursor, cursor + 4))
				cursor = cursor + 4
				table.insert(consts,dump:sub(cursor, cursor+off))
				cursor = cursor + off
			else
				-- Pass on nil
			end
		end
		return consts
	end

	local ok = pcall(function() string.dump(func) end)
	if not ok then return {"..."} end

	local dump = string.dump(func)
	local ret = _get_constants(dump)

	local function escape(str)
		return string.join(" ", str:split())
	end

	for _, c in ipairs(ret) do
		if escape(c) == "__help__" then
			return ret[_+1]
		end
	end
end

cmds = {}
if addhook then addhook("say", "cmds.sayhook", 10) end
function cmds.sayhook(id, text)
	local __return__ = 0
	for cat,prefix in pairs(cmds.prefixes) do
		local card = text:sub(1, #prefix)
		if card == prefix then
			text = text:sub(#prefix+1)
			local args = text:qsplit()
			local cmd = args[1]:trim()
			text = text:sub(#cmd+2)
			table.remove(args, 1)
			--Check that cmd is in category
			local category = cmds.categories[cat]
			if category:have(cmd) then
				--Check that there are sufficient parameters
				local group = category.commands[cmd]
				local func = group[1]
				local expected_args = group[2]
				local nargs = group[3]
				local realn = #args
				if table.find(args, "id") then
					--realn = realn - 1
				end
				if table.find(args, "text") then
					--realn = realn - 1
				end
				if realn < nargs then
					msg2(id, "Insufficient number of parameters for the command '"..cmd.."'")
					local vargs = ""
					if expected_args['...'] then vargs = ", [...]" end
					local eargs = copy(expected_args)
					if table.find(eargs,"id") then
						table.remove(eargs, table.find(eargs, "id"))
					end
					if table.find(eargs,"this") then
						table.remove(eargs, table.find(eargs, "this"))
					end
					while #eargs > nargs do
						table.remove(eargs)
					end
					msg2(id, "Need " .. nargs .. (expected_args['...'] or "") .. " arguments: " .. category.prefix .. cmd .. " <" .. string.join(", ",eargs) .. vargs .. ">")
					return 1
				end

				local this = {rank = group.rank}
				local determinant = category:preprocessing(id, cmd, text, this)
				this = determinant
				if type(this) ~= "table" then this = {} end
				this.rank = group.rank

				local builder = {}
				local cursor = 1
				for _,arg in ipairs(expected_args) do
					if arg == "id" then
						table.insert(builder, id)
					elseif arg == "this" then
						table.insert(builder, this)
					elseif arg == "text" then
						table.insert(builder, text)
						cursor = cursor + 0
					else
						table.insert(builder, args[cursor])
						cursor = cursor + 1
					end
				end
				if expected_args['...'] then
					for i=cursor, #args do
						table.insert(builder, args[i])
					end
				end

				if determinant then
					func(unpack(builder))
				end
				category:postprocessing(id, cmd, text, determinant)
			else
				--Display Command Not Found message... if enabled
				if category.not_found then
					msg2(id, string.format("252058050The command '%s' was not found.", cmd))
				end
			end
			__return__ = 1
		end
	end

	return __return__
end
cmds.categories = {}
cmds.prefixes = {}
cmds.category = function(category, prefix)
	local object = {}
	object.preprocessing = function(self, id, cmd, text)
		return true
	end
	object.prefix = prefix
	object.postprocessing = function(self, id, cmd, text)
		return true
	end
	object.commands = {}
	object.have = function(self, obj)
		if self.commands[obj] then return true else return false end
	end
	object.not_found = true
	object.rank = "default"

	setmetatable(object, {
		__newindex = function(self, k, v)
			if rawget(self, k) then
				if k == "postprocessing" or k == "preprocessing" then
					if type(v) ~= "function" then return end
				elseif k == "prefix" then
					if type(v) ~= "string" then return end
				elseif k == "not_found" then
					--
				elseif k == "rank" then
					if type(v) ~= "string" then v = tostring(v) end
				else
					return
				end
				rawset(self, k, v)
			end
			if type(v) ~= "function" then return end
			k = k:split("__")
			local args = string.join(" ",get_args(v)):split()
			local nargs = k[2] or num_args(v)
			if nargs:find("s") then
				args['...'] = "+"
				nargs = nargs:gsub("s","")
			end
			nargs = tonumber(nargs)
			if table.find(args, "id") and not k[2] then
				nargs = nargs - 1
			end
			if table.find(args, "this") and not k[2] then
				nargs = nargs - 1
			end
			if table.find(args, "text") then
				nargs = nargs
			end

			self.commands[k[1]] = {v, args, nargs, rank = self.rank}

			if help then
				local __help__ = get_help(v)
				if not __help__ then return end
				local prefix = self.prefix
				local fn = help.prefixes[prefix]
				fn{[k[1]]=__help__}
			end
		end,
		__index = function(self, k)
			if rawget(self, k) then return rawget(self, k) end
			local obj = {}
			setmetatable(obj, {__newindex = function(_, hook, v)
				--k is hook, v is function :split("__")[1]
				self[k] = v
				self.commands[k:split("__")[1]].rank = hook
			end})
			return obj
		end
	})
	cmds.prefixes[category] = prefix
	cmds.categories[category] = object
	return object
end

act = cmds.category("action", "!")

if help then
	function act.help(...)
		return
	end
end

function act.see(id, cmd)
	__help__ =
	[[
	-Usage:
	... !see <cmd> -> Shows the required (<argument>) and optional ([...])
	... arguments for a given command <cmd>.
	-Example:
	... !see see
	... ... Introspecting the action command 'see'
	... ... Need 1 arguments: !see <cmd>
	]]
	msg2(id, "250250250!see '%s'"%cmd)
	for _, category in pairs(cmds.categories) do
		if category:have(cmd) then
			local group = category.commands[cmd]
			local expected_args = group[2]
			local nargs = group[3]
			msg2(id, "250250250Introspecting the " .. _ .. " command '"..cmd.."'")
			local vargs = ""
			if expected_args['...'] then vargs = ", [...]" end
			local eargs = copy(expected_args)
			if table.find(eargs,"id") then
				table.remove(eargs, table.find(eargs, "id"))
			end
			while #eargs > nargs do
				table.remove(eargs)
			end
			msg2(id, "250250250Need " .. nargs .. (expected_args['...'] or "") .. " arguments: " .. category.prefix .. cmd .. " <" .. string.join(", ",eargs) .. vargs .. ">")
			--return
		end
	end
end

--~ function act.test(id)
--~ 	__help__ =
--~ 	[[
--~ 	-Desc:
--~ 	... Dummy command
--~ 	-Usage:
--~ 	... !test -> Displays TEST
--~ 	]]
--~ 	msg2(id, "TEST")
--~ end

--~ if help then
--~ 	local __gmt = getmetatable(_G) or {}
--~ 	local oni = __gmt.newindex or function(self, key, val)
--~ 		rawset(self, key, val)
--~ 	end
--~ 	__gmt.newindex = function(self, key, val)
--~ 		if type(oni) == "function" then
--~ 			oni(self, key, val)
--~ 		end
--~ 		if key == "__help__" then
--~ 			rawset(self, key, nil)
--~ 		end
--~ 	end
--~ end

