import "string"
import "menu"

help = {}

help.registry = {}
help.full = {}
help.half = {}
help.categories = {}
help.prefixes = {}

help.menu = Menu("Help")
help.menu.hooks._ = function(id, sel, m)
	local cmd = m.elements[sel]
	return help.default(id, cmd, help.registry[help.half[cmd]])

end
--help.register("sv_new", 'text") -> !sv_new
--help.register{sv_new="text",...} -> !sv_new

function help.make(category, prefix, default)
	local object = {}
	object.category = category or "generic"
	object.prefix = prefix or "!"

	if type(object.category) ~= "string" then return end
	if type(object.prefix) ~= "string" then return end

	if not help.categories[object.category] then help.categories[object.category] = {} end
	setmetatable(object,{__call = function(self,...)
		local args = {...}
		if type(args[1]) == "string" then
			args = {[args[1]] = args[2]}
		else
			args = args[1]
		end
		for k,pair in pairs(args) do
			local cmd = self.prefix .. k
			local content = pair
			help.registry[k] = content
			help.full[k] = cmd
			help.full[cmd] = k
			help.half[cmd] = k
			help.half[k] = k
			help.categories[self.category][k] = self.prefix
			help.menu:add(cmd)
			table.sort(help.menu.elements)
		end
	end})
	help.prefixes[object.prefix] = object
	return object
end

help.register = help.make()
help.admin = help.make("admin","@")

help.primary = "032178170"
help.secondary = "250250250"
help.default = function(id, cmd, content)
	msg2(id, help.primary .. "Command '" .. help.full[cmd] .. "':")
	if type(content) == "function" then
		local c = content(id)
		content = c or ""
	else
		content = tostring(content)
	end
	for _,line in ipairs(content:split("\n")) do
		msg2(id, help.secondary .. "    "..line:trim())
	end
end

if addhook then addhook("say", "hook_say_help") end
function hook_say_help(id, text)
	local args = text:split()
	if args[1] == "!help" or args[1] == "?" then
		local cmd = args[2]
		if not cmd or not #cmd then
			help.menu(id)
			return 1
		end
		if help.registry[cmd] then
			help.default(id, cmd, help.registry[help.half[cmd]])
			return 1
		end
		local m
		for k,v in pairs(help.registry) do
			if k:match(cmd) then
				if not m then
					m = Menu(string.format("Help '%s'", cmd))
					m.hooks._ = help.menu.hooks._
				end
				m:add(help.full[k])
			end
		end
		if m then m(id); return 1 end
		msg2(id, help.secondary .. "The command '" .. cmd .. "' was not found.")
		return 1
	end
end


help.register{
				help =
				[[
					-Aliases: ?
					-Usages:
					... !help -> Brings up the Help menu with all registered commands
					... !help <command> -> Brings up a help message of the command
					... !help <expr> -> Brings up a menu of all commands that contains expr
				]],
			}

import "cmds"
