import "string"

--[[Class]]Menu = {
	id = 0,
	static = {},
	decode = function(str)
		local locker = {}
		for element in str:gmatch("[^,]+") do
			table.insert(locker,element)
		end
		return locker
	end
}

setmetatable(Menu, {__call = function(self, title, ...)
	title = title or "Menu#"..tostring(Menu.id)
	local args = {...}

	local elements = {}
	for i,pair in ipairs(args) do
		local current
		if type(pair) == "string" then
			current = Menu.decode(pair)
		else
			current = pair
		end
		for _i,_pair in ipairs(current) do
			table.insert(elements, _pair)
		end
	end
	local object = {
		title = title,
		id = Menu.id,
		elements = elements,
		page = {},
		next = function(self, id)
			if not self.page[id] then self.page[id] = 1 end
			if self.page[id] < self.pages then
				self.page[id] = self.page[id] + 1
			end
			self(id)
		end,
		prev = function(self, id)
			if not self.page[id] then self.page[id] = 1 end
			if self.page[id] > 1 then
				self.page[id] = self.page[id] - 1
			end
			self(id)
		end,
		hooks = {},
		add = function(self, ...)
			local args = {...}
			local elements = {}
			for i,pair in ipairs(args) do
				local current
				if type(pair) == "string" then
					current = Menu.decode(pair)
				else
					current = pair
				end
				for _i,_pair in ipairs(current) do
					table.insert(elements, _pair)
				end
			end
			for _,e in ipairs(elements) do
				table.insert(self.elements,e)
			end
			self.pages = math.ceil(#self.elements/7)
		end,
	}
	setmetatable(object, {__call=function(self, id)
		--Render
		if not self.page[id] then self.page[id] = 1 end
		local serial
		local del = ","
		if self.pages == 1 then
			serial = self.title .. ","
			serial = serial .. del:join(self.elements)
		else
			serial = self.title .. string.format(" -=*Page %s*=-", self.page[id]) .. ","
			if self.page[id] > 1 then
				serial = serial .. "Previous ("..tostring(self.page[id]-1).."),"
			else
				serial = serial .. ","
			end
			local frame = {}
			for i=7*(self.page[id]-1)+1, self.page[id]*7 do table.insert(frame, self.elements[i]) end
			serial = serial .. del:join(frame)
			if self.page[id] < self.pages then
				serial = serial .. ",Next ("..tostring(self.page[id]+1).."),"
			end
		end
		menu(id, serial)
	end})
	if #object.elements <= 9 then
		object.pages = 1
	else
		object.pages = math.ceil(#object.elements/7)
	end
	setmetatable(object.hooks, {__newindex = function(self, k, v)
		--[[--
		Selection method
		menu.hooks.*
		menu.hooks.*_*
		menu.hooks._*
		menu.hooks._
		--]]--
		rawset(self,k,v)
	end})
	Menu.id = Menu.id + 1
	table.insert(Menu.static, object)
	return object
end})

if addhook then addhook("menu", "hook_menus") end
function hook_menus(id, title, selection)
	--Look for matching menus
	title = title:gsub(" %-=%*Page %d+%*=%-", "")

	for _,m in ipairs(Menu.static) do
		if title==m.title then
			local button
			local sel
			if m.pages == 1 then
				sel = selection
				button = m.elements[selection]
			else
				if selection == 1 then
					return m:prev(id)
				elseif selection == 9 then
					return m:next(id)
				end
				sel = 7*(m.page[id]-1)+selection-1
				button = m.elements[7*(m.page[id]-1)+selection-1]
			end
			local ordinal = 1
			for i,b in ipairs(m.elements) do
				if b == button and i<sel then
					ordinal = ordinal + 1
				end
			end

			--*_*
			local key
			key = button.."_"..tostring(ordinal)
			if m.hooks[key] then
				--function(id)
				return m.hooks[key](id)
			end
			--*
			key = button
			if m.hooks[key] then
				--function(id)
				return m.hooks[key](id)
			end
			--_*
			key = "_"..tostring(sel)
			if m.hooks[key] then
				--function(id)
				return m.hooks[key](id)
			end
			--_
			key = "_"
			if m.hooks[key] then
				--function(self, id, sel)
				return m.hooks[key](id, sel, m)
			end
			break
		end
	end
end
